#include "gfx.h"
#include "renderer_api.h"

#include "runtime/config.h"
#include "runtime/handles.h"
#include "runtime/hashing.h"
#include "runtime/mem_arena.h"
#include "runtime/threading.h"

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

RT_CVAR_I(rt_MaxFramegraphs, "Maximum number of framegraphs. Default 16", 16);

#define RT_FRAMEGRAPH_MAX_PASSES         32
#define RT_FRAMEGRAPH_MAX_RENDER_TARGETS 32
#define RT_RENDERPASS_MAX_READS          8
#define RT_RENDERPASS_MAX_WRITES         8

typedef struct {
    rt_render_target_id id;
    rt_pixel_format format;
    unsigned int width;
    unsigned int height;
    unsigned int sample_count;
    rt_gpu_semaphore_handle semaphore;
    rt_render_target_handle api_render_target;
} rt_render_target;

typedef struct {
    rt_render_pass_id id;
    rt_render_pass_type type;
    const char *name;
    int execution_level;

    bool reads_swapchain;
    bool writes_swapchain;

    unsigned int read_count;
    unsigned int write_count;
    rt_render_pass_bind_fns bound_fns;
    rt_render_target_read reads[RT_RENDERPASS_MAX_READS];
    rt_render_target_write writes[RT_RENDERPASS_MAX_WRITES];
} rt_render_pass;

struct rt_framegraph_s {
    uint32_t pass_count;
    uint32_t render_target_count;

    rt_framegraph *next_free;

    rt_render_pass passes[RT_FRAMEGRAPH_MAX_PASSES];

    rt_render_target render_targets[RT_FRAMEGRAPH_MAX_RENDER_TARGETS];
};

static rt_framegraph *_framegraphs;
static rt_framegraph *_first_free;
static rt_mutex *_free_list_lock;

#define NAMES_CAPACITY 512
static char _name_buffer[512];
static char *_name_next;
static rt_mutex *_name_lock;

static void ReturnFrameGraph(rt_framegraph *framegraph) {
    rtLockMutex(_free_list_lock);
    framegraph->next_free = _first_free;
    _first_free           = framegraph;
    rtUnlockMutex(_free_list_lock);
}

rt_result InitFramegraphManager(void) {
    _free_list_lock = rtCreateMutex();
    if (!_free_list_lock)
        return RT_UNKNOWN_ERROR;
    _name_lock = rtCreateMutex();
    if (!_name_lock) {
        rtDestroyMutex(_free_list_lock);
        return RT_UNKNOWN_ERROR;
    }
    _framegraphs = calloc((size_t)rt_MaxFramegraphs.i, sizeof(rt_framegraph));
    if (!_framegraphs)
        return RT_OUT_OF_MEMORY;
    for (int i = 0; i < rt_MaxFramegraphs.i; ++i)
        _framegraphs[i].next_free = (i < rt_MaxFramegraphs.i - 1) ? &_framegraphs[i + 1] : NULL;
    _first_free = &_framegraphs[0];
    _name_next  = &_name_buffer[0];
    return RT_SUCCESS;
}

void ShutdownFramegraphManager(void) {
    free(_framegraphs);
    rtDestroyMutex(_free_list_lock);
    rtDestroyMutex(_name_lock);
}

typedef struct {
    unsigned int dependency_count;
    int execution_level;
} rt_pass_construct;

static int CompareRenderPassExecutionLevels(const void *a, const void *b) {
    const rt_render_pass *pass_a = a, *pass_b = b;
    return pass_a->execution_level - pass_b->execution_level;
}

static bool
CreateRenderPasses(rt_framegraph *graph, const rt_framegraph_info *info, rt_arena *arena) {
    uint32_t render_pass_count = info->render_pass_count;

    bool result = false;

    /* Pass A depends on pass B, if:
     * B preceeds A in the list of render passes AND
     * B writes to a render target that A reads from. */
    bool *dependency_matrix =
        rtArenaPushZero(arena, render_pass_count * render_pass_count * sizeof(bool));
    if (!dependency_matrix) {
        rtLog("GFX",
              "Not enough memory to allocate a %ux%u dependency matrix.",
              render_pass_count,
              render_pass_count);
        goto out;
    }
    /* Checks if pass "dependent_idx" depends on pass "dependency_idx" */
#define PASS_DEPENDS(dependent_idx, dependency_idx)                                                \
    dependency_matrix[(dependency_idx)*render_pass_count + (dependent_idx)]

    rt_pass_construct *construct_passes =
        RT_ARENA_PUSH_ARRAY_ZERO(arena, rt_pass_construct, render_pass_count);
    if (!construct_passes) {
        rtLog("GFX",
              "Not enough memory to allocate construction information for %u passes.",
              render_pass_count);
        goto out;
    }

    const rt_render_pass_info *pass_info = rtResolveConstRelptr(&info->render_passes);
    for (uint32_t i = 0; i < render_pass_count; ++i) {
        construct_passes[i].execution_level = -1; /* not scheduled yet */
        const rt_render_target_write *writes_i =
            rtResolveConstRelptr(&pass_info[i].write_render_targets);
        for (uint32_t j = i + 1; j < render_pass_count; ++j) {
            const rt_render_target_read *reads_j =
                rtResolveConstRelptr(&pass_info[j].read_render_targets);
            bool depends = false;
            for (uint32_t read_idx = 0; read_idx < pass_info[j].read_render_target_count;
                 ++read_idx) {
                for (uint32_t write_idx = 0; write_idx < pass_info[i].write_render_target_count;
                     ++write_idx) {
                    if (writes_i[write_idx].render_target == reads_j[read_idx].render_target)
                        depends = true;
                }
            }
            PASS_DEPENDS(j, i) = depends;
            if (depends)
                ++construct_passes[j].dependency_count;
        }
    }

    /* Pass A can be executed concurrently with pass B if:
     * 1. A and B don't write to the same render target AND
     * 2. A's dependencies and B's dependencies have finished executing. */

    /* We can have at most render_pass_count execution levels */
    uint32_t *level_passes = RT_ARENA_PUSH_ARRAY_ZERO(arena, uint32_t, render_pass_count);
    if (!level_passes) {
        rtLog("GFX", "Failed to allocate a temporary array for constructing execution levels.");
        goto out;
    }
    uint32_t unscheduled_passes = render_pass_count;
    for (int level = 0; level < (int)render_pass_count; ++level) {
        unsigned int level_pass_count = 0;
        for (uint32_t i = 0; i < render_pass_count; ++i) {
            if (construct_passes[i].execution_level == -1 &&
                construct_passes[i].dependency_count == 0) {

                /* Check that no writes conflict */
                bool write_conflict = false;
                const rt_render_target_write *writes_i =
                    rtResolveConstRelptr(&pass_info[i].write_render_targets);
                for (unsigned int j = 0; j < level_pass_count; ++j) {
                    uint32_t pass_idx = level_passes[i];
                    const rt_render_target_write *pass_writes =
                        rtResolveConstRelptr(&pass_info[pass_idx].write_render_targets);
                    for (uint32_t k = 0; k < pass_info[i].write_render_target_count; ++k) {
                        for (uint32_t l = 0; l < pass_info[pass_idx].write_render_target_count;
                             ++l) {
                            if (writes_i[k].render_target == pass_writes[l].render_target) {
                                write_conflict = true;
                                break;
                            }
                        }
                        if (write_conflict)
                            break;
                    }
                    if (write_conflict)
                        break;
                }
                if (!write_conflict) {
                    RT_ASSERT(level_pass_count < render_pass_count, "");
                    level_passes[level_pass_count++]    = i;
                    construct_passes[i].execution_level = level;
                }
            }
        }
        if (level_pass_count == 0) {
            rtLog("GFX", "Failed to compute a valid schedule for the provided framegraph.");
            goto out;
        }
        /* level passes now contains the passes we can execute concurrently.
         * Decrement dependency count for all passes that depend on a pass in this level */

        for (uint32_t i = 0; i < level_pass_count; ++i) {
            for (uint32_t j = 0; j < render_pass_count; ++j) {
                if (PASS_DEPENDS(j, level_passes[i]))
                    --construct_passes[j].dependency_count;
            }
        }

        unscheduled_passes -= level_pass_count;
        if (unscheduled_passes == 0)
            break;
    }
    RT_ASSERT(unscheduled_passes == 0, "Did not schedule all passes");
    /* Construct passes now contains the "execution level" for each pass.
     * We execute passes in that order, those with the same execution level can be executed
     * concurrently. */

    graph->pass_count = render_pass_count;
    for (uint32_t i = 0; i < render_pass_count; ++i) {
        graph->passes[i].execution_level = construct_passes[i].execution_level;
        const rt_render_target_write *writes =
            rtResolveConstRelptr(&pass_info[i].write_render_targets);
        const rt_render_target_read *reads =
            rtResolveConstRelptr(&pass_info[i].read_render_targets);
        memcpy(graph->passes[i].writes,
               writes,
               pass_info[i].write_render_target_count * sizeof(rt_render_target_write));
        memcpy(graph->passes[i].reads,
               reads,
               pass_info[i].read_render_target_count * sizeof(rt_render_target_read));
        graph->passes[i].write_count = pass_info[i].write_render_target_count;
        graph->passes[i].read_count  = pass_info[i].read_render_target_count;
        graph->passes[i].id          = pass_info[i].id;
        graph->passes[i].type        = pass_info[i].type;
        graph->passes[i].name        = NULL;

        graph->passes[i].reads_swapchain  = false;
        graph->passes[i].writes_swapchain = false;
        const rt_render_target_info *rts  = rtResolveConstRelptr(&info->render_targets);
        for (unsigned int j = 0; j < graph->passes[i].read_count; ++j) {
            rt_render_target_id rt = graph->passes[i].reads[j].render_target;
            for (unsigned int k = 0; k < info->render_target_count; ++k) {
                if (rts[k].id == rt && rts[k].width == RT_RENDER_TARGET_SIZE_SWAPCHAIN &&
                    rts[k].height == RT_RENDER_TARGET_SIZE_SWAPCHAIN &&
                    rts[k].format == RT_PIXEL_FORMAT_SWAPCHAIN) {
                    graph->passes[i].reads_swapchain = true;
                }
            }
        }
        for (unsigned int j = 0; j < graph->passes[i].write_count; ++j) {
            rt_render_target_id rt = graph->passes[i].writes[j].render_target;
            for (unsigned int k = 0; k < info->render_target_count; ++k) {
                if (rts[k].id == rt && rts[k].width == RT_RENDER_TARGET_SIZE_SWAPCHAIN &&
                    rts[k].height == RT_RENDER_TARGET_SIZE_SWAPCHAIN &&
                    rts[k].format == RT_PIXEL_FORMAT_SWAPCHAIN) {
                    graph->passes[i].writes_swapchain = true;
                }
            }
        }

        const char *name = rtResolveConstRelptr(&pass_info[i].name);
        if (name) {
            size_t name_strlen = strlen(name);
            if (name_strlen + 1 == pass_info[i].name_len) {
                rtLockMutex(_name_lock);
                ptrdiff_t name_off = _name_next - _name_buffer;
                if ((name_off + pass_info[i].name_len) < NAMES_CAPACITY) {
                    char *dst_name = _name_next;
                    memcpy(dst_name, name, pass_info[i].name_len);
                    _name_next += pass_info[i].name_len;
                    graph->passes[i].name = dst_name;
                } else {
                    rtLog("GFX", "Ran out of storage for debug name %s", name);
                }
                rtUnlockMutex(_name_lock);
            } else {
                rtLog("GFX", "Declared name-length for pass %u does not match strlen()");
            }
        }
    }

    /* Sort by execution level */
    qsort(graph->passes,
          render_pass_count,
          sizeof(rt_render_pass),
          CompareRenderPassExecutionLevels);
    result = true;
out:
    return result;
#undef PASS_DEPENDS
}

static bool
CreateRenderTargets(rt_framegraph *graph, const rt_framegraph_info *info, rt_arena *arena) {
    bool result = false;

    /* TODO(Kevin): determine aliasing opportunities */
    const rt_render_target_info *render_targets = rtResolveConstRelptr(&info->render_targets);
    for (uint32_t i = 0; i < info->render_target_count; ++i) {
        graph->render_targets[i].id           = render_targets[i].id;
        graph->render_targets[i].format       = render_targets[i].format;
        graph->render_targets[i].width        = render_targets[i].width;
        graph->render_targets[i].height       = render_targets[i].height;
        graph->render_targets[i].sample_count = render_targets[i].sample_count;

        rt_gpu_semaphore_info sem_info = {
            .initial_value = 0,
            .name          = NULL,
        };
        g_renderer.CreateSemaphores(1, &sem_info, &graph->render_targets[i].semaphore);

        if (graph->render_targets[i].width != RT_RENDER_TARGET_SIZE_SWAPCHAIN ||
            graph->render_targets[i].height != RT_RENDER_TARGET_SIZE_SWAPCHAIN ||
            graph->render_targets[i].format != RT_PIXEL_FORMAT_SWAPCHAIN) {

            graph->render_targets[i].api_render_target =
                g_renderer.CreateRenderTarget(&render_targets[i]);
        } else {
            graph->render_targets[i].api_render_target = g_renderer.GetSwapchainRenderTarget();
        }
        if (!RT_IS_HANDLE_VALID(graph->render_targets[i].api_render_target)) {
            rtReportError("GFX", "Failed to create render target %u of framegraph.", i);
            for (uint32_t j = 0; j < i; ++j)
                g_renderer.DestroyRenderTarget(graph->render_targets[j].api_render_target);
            goto out;
        }
    }
    graph->render_target_count = info->render_target_count;

    result = true;
out:
    return result;
}

static bool ValidateInfo(const rt_framegraph_info *info) {
    if (info->render_pass_count > RT_FRAMEGRAPH_MAX_PASSES) {
        rtReportError("GFX",
                      "Framegraph has too many passes: %u (maximum allowed is %u)",
                      info->render_pass_count,
                      RT_FRAMEGRAPH_MAX_PASSES);
        return false;
    }
    if (info->render_target_count > RT_FRAMEGRAPH_MAX_RENDER_TARGETS) {
        rtReportError("GFX",
                      "Framegraph has too many render targets: %u (maximum allowed is %u)",
                      info->render_target_count,
                      RT_FRAMEGRAPH_MAX_RENDER_TARGETS);
        return false;
    }

    const rt_render_target_info *render_targets = rtResolveConstRelptr(&info->render_targets);
    for (uint32_t i = 0; i < info->render_target_count; ++i) {
        if (render_targets[i].id == 0) {
            rtReportError("GFX", "Framegraph render target %u has invalid id 0", i);
            return false;
        } else if ((render_targets[i].width == RT_RENDER_TARGET_SIZE_SWAPCHAIN ||
                    render_targets[i].height == RT_RENDER_TARGET_SIZE_SWAPCHAIN) &&
                   (render_targets[i].width != render_targets[i].height)) {
            rtReportError("GFX",
                          "Framegraph render target %u: If width or height is set to "
                          "SWAPCHAIN, both values must be set to SWAPCHAIN.",
                          i);
            return false;
        } else if (render_targets[i].format >= RT_PIXEL_FORMAT_count) {
            rtReportError("GFX",
                          "Framegraph render target %u format is outside the allowed range.",
                          i);
            return false;
        }
    }

    const rt_render_pass_info *passes = rtResolveConstRelptr(&info->render_passes);
    for (uint32_t i = 0; i < info->render_pass_count; ++i) {
        if (passes[i].id == 0) {
            rtReportError("GFX", "Framegraph pass %u has invalid id 0", i);
            return false;
        } else if (passes[i].read_render_target_count > RT_RENDERPASS_MAX_READS) {
            rtReportError(
                "GFX",
                "Framegraph pass %u reads too many rendertargets: %u (maximum allowed is %u)",
                i,
                passes[i].read_render_target_count,
                RT_RENDERPASS_MAX_READS);
            return false;
        } else if (passes[i].write_render_target_count > RT_RENDERPASS_MAX_WRITES) {
            rtReportError(
                "GFX",
                "Framegraph pass %u writes too many rendertargets: %u (maximum allowed is %u)",
                i,
                passes[i].write_render_target_count,
                RT_RENDERPASS_MAX_WRITES);
            return false;
        }
    }

    return true;
}

RT_DLLEXPORT rt_framegraph *rtCreateFramegraph(const rt_framegraph_info *info) {
    if (!ValidateInfo(info)) {
        return NULL;
    }

    rt_temp_arena temp = rtGetTemporaryArena(NULL, 0);
    if (!temp.arena) {
        rtReportError("GFX", "Failed to acquire a temporary arena for constructing a framegraph");
        return NULL;
    }

    rt_framegraph *graph = NULL;
    /* Acquire a unused framegraph */
    rtLockMutex(_free_list_lock);
    graph = _first_free;
    if (graph)
        _first_free = graph->next_free;
    rtUnlockMutex(_free_list_lock);
    if (!graph)
        goto out;
    memset(graph, 0, sizeof(*graph));

    if (!CreateRenderPasses(graph, info, temp.arena)) {
        ReturnFrameGraph(graph);
        graph = NULL;
        goto out;
    }

    if (!CreateRenderTargets(graph, info, temp.arena)) {
        ReturnFrameGraph(graph);
        graph = NULL;
        goto out;
    }

out:
    rtReturnTemporaryArena(temp);
    return graph;
}

RT_DLLEXPORT void rtDestroyFramegraph(rt_framegraph *framegraph) {
    ReturnFrameGraph(framegraph);
}

RT_DLLEXPORT void rtBindRenderPass(rt_framegraph *framegraph,
                                   rt_render_pass_id id,
                                   const rt_render_pass_bind_fns *bind_fns) {
    for (uint32_t i = 0; i < framegraph->pass_count; ++i) {
        if (framegraph->passes[i].id == id) {
            if (framegraph->passes[i].bound_fns.Execute)
                rtLog("GFX", "Rebound pass %x to new functions", id);
            framegraph->passes[i].bound_fns = *bind_fns;
            return;
        }
    }
    rtLog("GFX", "Tried to bind functions to unknown render pass %x", id);
}

static bool IsDepthFormat(rt_pixel_format format) {
    return format == RT_PIXEL_FORMAT_DEPTH32 || format == RT_PIXEL_FORMAT_DEPTH24_STENCIL8;
}

static rt_render_target *GetRenderTarget(rt_framegraph *framegraph, rt_render_target_id id) {
    for (uint32_t i = 0; i < framegraph->render_target_count; ++i) {
        if (framegraph->render_targets[i].id == id)
            return &framegraph->render_targets[i];
    }
    return NULL;
}

static void
BeginGraphicsPass(rt_framegraph *framegraph, uint32_t pass_idx, rt_command_buffer_handle cmdbuf) {
    const rt_render_target_write *writes = framegraph->passes[pass_idx].writes;
    const rt_render_target_read *reads   = framegraph->passes[pass_idx].reads;
    uint32_t write_count                 = framegraph->passes[pass_idx].write_count;
    uint32_t read_count                  = framegraph->passes[pass_idx].read_count;

    /* Convert reads and writes into the pass begin info for the renderer */
    rt_cmd_begin_pass_info begin_info;
    memset(&begin_info, 0, sizeof(begin_info));

    /* All written render targets need to have the same size */
    if (write_count > 0) {
        rt_render_target *rt = GetRenderTarget(framegraph, writes[0].render_target);
        RT_ASSERT(rt != NULL, "Invalid render target in pass write.");
        begin_info.render_area = (rt_rect2i){
            .offset = {             0,               0},
            .size   = {.x = rt->width, .y = rt->height}
        };
    }

    for (uint32_t i = 0; i < write_count; ++i) {
        rt_render_target *rt = GetRenderTarget(framegraph, writes[i].render_target);
        RT_ASSERT(rt != NULL, "Invalid render target in pass write.");

        g_renderer.CmdTransitionRenderTarget(cmdbuf,
                                             rt->api_render_target,
                                             RT_RENDER_TARGET_STATE_ATTACHMENT);

        if (!IsDepthFormat(rt->format)) {
            /* Add as color buffer */
            uint32_t cbidx = begin_info.color_buffer_count;
            RT_ASSERT(cbidx < 4, "Maximum of 4 colorbuffers per pass exceeded.");
            begin_info.color_buffers[cbidx] = rt->api_render_target;
            if ((writes[i].flags & RT_RENDER_TARGET_WRITE_CLEAR) != 0) {
                begin_info.color_buffer_loads[cbidx]              = RT_PASS_LOAD_MODE_CLEAR;
                begin_info.color_buffer_clear_values[cbidx].color = writes[i].clear.color;
            } else {
                begin_info.color_buffer_loads[cbidx] = RT_PASS_LOAD_MODE_LOAD;
            }
            if ((writes[i].flags & RT_RENDER_TARGET_WRITE_DISCARD) != 0) {
                begin_info.color_buffer_writes[cbidx] = RT_PASS_WRITE_MODE_DISCARD;
            } else {
                begin_info.color_buffer_writes[cbidx] = RT_PASS_WRITE_MODE_STORE;
            }
            ++begin_info.color_buffer_count;
        } else {
            /* Add as depth buffer*/
            RT_ASSERT(!RT_IS_HANDLE_VALID(begin_info.depth_stencil_buffer),
                      "Only one depth/stencil buffer can be set!");
            begin_info.depth_stencil_buffer = rt->api_render_target;
            if ((writes[i].flags & RT_RENDER_TARGET_WRITE_CLEAR) != 0) {
                begin_info.depth_stencil_buffer_load = RT_PASS_LOAD_MODE_CLEAR;
                begin_info.depth_stencil_buffer_clear_value.depth_stencil.depth =
                    writes[i].clear.depth_stencil.depth;
                begin_info.depth_stencil_buffer_clear_value.depth_stencil.stencil =
                    writes[i].clear.depth_stencil.stencil;
            } else {
                begin_info.depth_stencil_buffer_load = RT_PASS_LOAD_MODE_LOAD;
            }
            if ((writes[i].flags & RT_RENDER_TARGET_WRITE_DISCARD) != 0) {
                begin_info.depth_stencil_buffer_write = RT_PASS_WRITE_MODE_DISCARD;
            } else {
                begin_info.depth_stencil_buffer_write = RT_PASS_WRITE_MODE_STORE;
            }
        }
    }

    for (uint32_t i = 0; i < read_count; ++i) {
        rt_render_target *rt = GetRenderTarget(framegraph, reads[i].render_target);
        RT_ASSERT(rt != NULL, "Invalid render target in pass read.");
        /* We need to transition the render target */

        switch (reads[i].mode) {
        case RT_RENDER_TARGET_READ_SAMPLED:
            g_renderer.CmdTransitionRenderTarget(cmdbuf,
                                                 rt->api_render_target,
                                                 RT_RENDER_TARGET_STATE_SAMPLED_IMAGE);
            break;
        case RT_RENDER_TARGET_READ_DIRECT:
            g_renderer.CmdTransitionRenderTarget(cmdbuf,
                                                 rt->api_render_target,
                                                 RT_RENDER_TARGET_STATE_STORAGE_IMAGE);
            break;
        default:
            RT_ASSERT(0, "Invalid render target read mode");
        }
    }

    g_renderer.CmdBeginPass(cmdbuf, &begin_info);
}

static void
BeginComputePass(rt_framegraph *framegraph, uint32_t pass_idx, rt_command_buffer_handle cmdbuf) {
    const rt_render_target_write *writes = framegraph->passes[pass_idx].writes;
    const rt_render_target_read *reads   = framegraph->passes[pass_idx].reads;
    uint32_t write_count                 = framegraph->passes[pass_idx].write_count;
    uint32_t read_count                  = framegraph->passes[pass_idx].read_count;

    for (uint32_t i = 0; i < write_count; ++i) {
        rt_render_target *rt = GetRenderTarget(framegraph, writes[i].render_target);
        RT_ASSERT(rt != NULL, "Invalid render target in pass write.");
        g_renderer.CmdTransitionRenderTarget(cmdbuf,
                                             rt->api_render_target,
                                             RT_RENDER_TARGET_STATE_STORAGE_IMAGE);
    }

    for (uint32_t i = 0; i < read_count; ++i) {
        rt_render_target *rt = GetRenderTarget(framegraph, reads[i].render_target);
        RT_ASSERT(rt != NULL, "Invalid render target in pass read.");
        /* We need to transition the render target */

        switch (reads[i].mode) {
        case RT_RENDER_TARGET_READ_SAMPLED:
            g_renderer.CmdTransitionRenderTarget(cmdbuf,
                                                 rt->api_render_target,
                                                 RT_RENDER_TARGET_STATE_SAMPLED_IMAGE);
            break;
        case RT_RENDER_TARGET_READ_DIRECT:
            g_renderer.CmdTransitionRenderTarget(cmdbuf,
                                                 rt->api_render_target,
                                                 RT_RENDER_TARGET_STATE_STORAGE_IMAGE);
            break;
        default:
            RT_ASSERT(0, "Invalid render target read mode");
        }
    }
}

RT_DLLEXPORT void rtExecuteFramegraph(rt_framegraph *framegraph, unsigned int frame_id) {
    int execution_level  = framegraph->passes[0].execution_level;
    uint32_t level_start = 0;

    rt_gpu_semaphore_handle swapchain_available = g_renderer.GetSwapchainAvailableSemaphore();
    rt_gpu_semaphore_handle render_finished     = g_renderer.GetRenderFinishedSemaphore();

    rt_temp_arena temp = rtGetTemporaryArena(NULL, 0);
    if (!temp.arena) {
        rtLog("GFX", "Unable to execute framegraph because no temporary arena is available.");
        return;
    }

    /* Find the last pass that writes to the swapchain */
    uint32_t last_swapchain_write = 0;
    for (uint32_t i = framegraph->pass_count - 1; i > 0; --i) {
        if (framegraph->passes[i].writes_swapchain) {
            last_swapchain_write = i;
            break;
        }
    }
    /* Find the first pass that reads the swapchain 0*/
    uint32_t first_swapchain_read = 0;
    for (uint32_t i = 0; framegraph->pass_count; ++i) {
        if (framegraph->passes[i].reads_swapchain || framegraph->passes[i].writes_swapchain) {
            first_swapchain_read = i;
            break;
        }
    }

    /* Overflows after ~4.871x10^7 years */
    uint64_t signal_value_base = (uint64_t)frame_id * 100;

    for (uint32_t i = 0; i <= framegraph->pass_count && level_start < framegraph->pass_count; ++i) {
        if ((i == framegraph->pass_count) ||
            (framegraph->passes[i].execution_level > execution_level)) {

            rt_temp_arena level_temp = rtBeginTempArena(temp.arena);

            rt_gpu_semaphore_handle *graphics_wait_semaphores   = NULL;
            rt_gpu_semaphore_handle *graphics_signal_semaphores = NULL;
            uint64_t *graphics_wait_values                      = NULL;
            uint64_t *graphics_signal_values                    = NULL;
            rt_command_buffer_handle *graphics_command_buffers  = NULL;
            uint32_t graphics_command_buffer_count              = 0;
            uint32_t graphics_signal_semaphore_count            = 0;
            uint32_t graphics_wait_semaphore_count              = 0;

            rt_gpu_semaphore_handle *compute_wait_semaphores   = NULL;
            rt_gpu_semaphore_handle *compute_signal_semaphores = NULL;
            rt_command_buffer_handle *compute_command_buffers  = NULL;
            uint64_t *compute_wait_values                      = NULL;
            uint64_t *compute_signal_values                    = NULL;
            uint32_t compute_command_buffer_count              = 0;
            uint32_t compute_signal_semaphore_count            = 0;
            uint32_t compute_wait_semaphore_count              = 0;

            /* Determine necessary array sizes */
            for (uint32_t pass_idx = level_start; pass_idx < i; ++pass_idx) {
                bool pass_bound = framegraph->passes[pass_idx].bound_fns.Prepare != NULL &&
                                  framegraph->passes[pass_idx].bound_fns.Execute != NULL &&
                                  framegraph->passes[pass_idx].bound_fns.Finalize != NULL;
                if (!pass_bound) {
                    rtLog("GFX",
                          "Framegraph pass %u (%x) is not bound to any function.",
                          pass_idx,
                          framegraph->passes[pass_idx].id);
                    continue;
                }
                bool is_graphics_pass =
                    framegraph->passes[pass_idx].type == RT_RENDER_PASS_TYPE_GRAPHICS;
                if (is_graphics_pass) {
                    graphics_wait_semaphore_count += framegraph->passes[pass_idx].read_count;
                    graphics_signal_semaphore_count += framegraph->passes[pass_idx].write_count;
                    if (framegraph->passes[pass_idx].reads_swapchain)
                        graphics_wait_semaphore_count += 1;
                    if (framegraph->passes[pass_idx].writes_swapchain)
                        graphics_signal_semaphore_count += 1;
                    ++graphics_command_buffer_count;
                } else {
                    compute_wait_semaphore_count += framegraph->passes[pass_idx].read_count;
                    compute_signal_semaphore_count += framegraph->passes[pass_idx].write_count;
                    if (framegraph->passes[pass_idx].reads_swapchain)
                        compute_wait_semaphore_count += 1;
                    if (framegraph->passes[pass_idx].writes_swapchain)
                        compute_signal_semaphore_count += 1;
                    ++compute_command_buffer_count;
                }
            }

            graphics_wait_semaphores = RT_ARENA_PUSH_ARRAY(level_temp.arena,
                                                           rt_gpu_semaphore_handle,
                                                           graphics_wait_semaphore_count);
            graphics_wait_values =
                RT_ARENA_PUSH_ARRAY(level_temp.arena, uint64_t, graphics_wait_semaphore_count);
            graphics_signal_semaphores = RT_ARENA_PUSH_ARRAY(level_temp.arena,
                                                             rt_gpu_semaphore_handle,
                                                             graphics_signal_semaphore_count);
            graphics_signal_values =
                RT_ARENA_PUSH_ARRAY(level_temp.arena, uint64_t, graphics_signal_semaphore_count);
            graphics_command_buffers        = RT_ARENA_PUSH_ARRAY(level_temp.arena,
                                                           rt_command_buffer_handle,
                                                           graphics_command_buffer_count);
            graphics_signal_semaphore_count = 0;
            graphics_wait_semaphore_count   = 0;
            graphics_command_buffer_count   = 0;

            compute_wait_semaphores = RT_ARENA_PUSH_ARRAY(level_temp.arena,
                                                          rt_gpu_semaphore_handle,
                                                          compute_wait_semaphore_count);
            compute_wait_values =
                RT_ARENA_PUSH_ARRAY(level_temp.arena, uint64_t, compute_wait_semaphore_count);
            compute_signal_semaphores = RT_ARENA_PUSH_ARRAY(level_temp.arena,
                                                            rt_gpu_semaphore_handle,
                                                            compute_signal_semaphore_count);
            compute_signal_values =
                RT_ARENA_PUSH_ARRAY(level_temp.arena, uint64_t, compute_signal_semaphore_count);
            compute_command_buffers        = RT_ARENA_PUSH_ARRAY(level_temp.arena,
                                                          rt_command_buffer_handle,
                                                          compute_command_buffer_count);
            compute_signal_semaphore_count = 0;
            compute_wait_semaphore_count   = 0;
            compute_command_buffer_count   = 0;

            /* Dispatch all passes in the current execution level */
            for (uint32_t pass_idx = level_start; pass_idx < i; ++pass_idx) {
                bool pass_bound = framegraph->passes[pass_idx].bound_fns.Prepare != NULL &&
                                  framegraph->passes[pass_idx].bound_fns.Execute != NULL &&
                                  framegraph->passes[pass_idx].bound_fns.Finalize != NULL;
                if (!pass_bound)
                    continue;

                // rt_render_pass_id id                 = framegraph->passes[pass_idx].id;
                const rt_render_target_write *writes = framegraph->passes[pass_idx].writes;
                const rt_render_target_read *reads   = framegraph->passes[pass_idx].reads;
                uint32_t write_count                 = framegraph->passes[pass_idx].write_count;
                uint32_t read_count                  = framegraph->passes[pass_idx].read_count;

                /* TODO(Kevin): Every one of these should be a job-dispatch*/

                bool is_graphics_pass =
                    framegraph->passes[pass_idx].type == RT_RENDER_PASS_TYPE_GRAPHICS;
                rt_command_buffer_handle cmdbuf;
                rt_alloc_command_buffer_info cmdbuf_alloc = {
                    .target_queue = is_graphics_pass ? RT_GRAPHICS_QUEUE : RT_COMPUTE_QUEUE,
                };
                if (g_renderer.AllocCommandBuffers(1, &cmdbuf_alloc, &cmdbuf) != RT_SUCCESS) {
                    rtLog("GFX",
                          "Failed to allocate a command buffer for framegraph pass %u (%x)",
                          pass_idx,
                          framegraph->passes[pass_idx].id);
                    continue;
                }

                if (is_graphics_pass) {
                    BeginGraphicsPass(framegraph, pass_idx, cmdbuf);
                } else {
                    BeginComputePass(framegraph, pass_idx, cmdbuf);
                }

                /*
                framegraph->passes[pass_idx].bound_fns.Prepare(id,
                                                               writes,
                                                               write_count,
                                                               reads,
                                                               read_count);
                framegraph->passes[pass_idx].bound_fns.Execute(id,
                                                               writes,
                                                               write_count,
                                                               reads,
                                                               read_count);
                framegraph->passes[pass_idx].bound_fns.Finalize(id,
                                                                writes,
                                                                write_count,
                                                                reads,
                                                                read_count);
                */
                if (is_graphics_pass) {
                    g_renderer.CmdEndPass(cmdbuf);
                }

                rt_gpu_semaphore_handle *wait_semaphores = NULL, *signal_semaphores = NULL;
                uint64_t *wait_values = NULL, *signal_values = NULL;
                rt_command_buffer_handle *command_buffers = NULL;
                uint32_t *wait_count = NULL, *signal_count = 0;
                uint32_t *command_buffer_count = NULL;
                if (is_graphics_pass) {
                    wait_semaphores      = graphics_wait_semaphores;
                    signal_semaphores    = graphics_signal_semaphores;
                    wait_values          = graphics_wait_values;
                    signal_values        = graphics_signal_values;
                    command_buffers      = graphics_command_buffers;
                    wait_count           = &graphics_wait_semaphore_count;
                    signal_count         = &graphics_signal_semaphore_count;
                    command_buffer_count = &graphics_command_buffer_count;
                } else {
                    wait_semaphores      = compute_wait_semaphores;
                    signal_semaphores    = compute_signal_semaphores;
                    wait_values          = compute_wait_values;
                    signal_values        = compute_signal_values;
                    command_buffers      = compute_command_buffers;
                    wait_count           = &compute_wait_semaphore_count;
                    signal_count         = &compute_signal_semaphore_count;
                    command_buffer_count = &compute_command_buffer_count;
                }

                for (uint32_t j = 0; j < read_count; ++j) {
                    rt_render_target *rt = GetRenderTarget(framegraph, reads[j].render_target);
                    wait_semaphores[*wait_count] = rt->semaphore;
                    wait_values[*wait_count]     = signal_value_base + execution_level;
                    *wait_count += 1;
                }
                for (uint32_t j = 0; j < write_count; ++j) {
                    rt_render_target *rt = GetRenderTarget(framegraph, writes[j].render_target);
                    signal_semaphores[*signal_count] = rt->semaphore;
                    signal_values[*signal_count]     = signal_value_base + execution_level + 1;
                    *signal_count += 1;
                }
                if (pass_idx == first_swapchain_read) {
                    wait_semaphores[*wait_count] = swapchain_available;
                    wait_values[*wait_count]     = 0;
                    *wait_count += 1;
                }
                if (pass_idx == last_swapchain_write) {
                    signal_semaphores[*signal_count] = render_finished;
                    signal_values[*signal_count]     = 0;
                    *signal_count += 1;
                }
                command_buffers[*command_buffer_count] = cmdbuf;
                *command_buffer_count += 1;
            }

            if (graphics_command_buffer_count > 0) {
                rt_submit_command_buffers_info submit = {
                    .command_buffers        = graphics_command_buffers,
                    .command_buffer_count   = graphics_command_buffer_count,
                    .signal_semaphores      = graphics_signal_semaphores,
                    .signal_values          = graphics_signal_values,
                    .signal_semaphore_count = graphics_signal_semaphore_count,
                    .wait_semaphores        = graphics_wait_semaphores,
                    .wait_values            = graphics_wait_values,
                    .wait_semaphore_count   = graphics_wait_semaphore_count,
                };
                g_renderer.SubmitCommandBuffers(RT_GRAPHICS_QUEUE, &submit);
            }

            if (compute_command_buffer_count > 0) {
                rt_submit_command_buffers_info submit = {
                    .command_buffers        = compute_command_buffers,
                    .command_buffer_count   = compute_command_buffer_count,
                    .signal_semaphores      = compute_signal_semaphores,
                    .signal_values          = compute_signal_values,
                    .signal_semaphore_count = compute_signal_semaphore_count,
                    .wait_semaphores        = compute_wait_semaphores,
                    .wait_values            = compute_wait_values,
                    .wait_semaphore_count   = compute_wait_semaphore_count,
                };
                g_renderer.SubmitCommandBuffers(RT_COMPUTE_QUEUE, &submit);
            }

            /* Start next level */
            level_start = i;
            if (i < framegraph->pass_count)
                execution_level = framegraph->passes[i].execution_level;
        }
    }
}

RT_DLLEXPORT rt_render_target_id rtCalculateRenderTargetID(const char *name, size_t len) {
    rt_render_target_id id = rtHashBytes32(name, len);
    if (id == 0)
        id = ~id;
    return id;
}

RT_DLLEXPORT rt_render_pass_id rtCalculateRenderPassID(const char *name, size_t len) {
    rt_render_pass_id id = rtHashBytes32(name, len);
    if (id == 0)
        id = ~id;
    return id;
}