#ifndef RT_GFX_BACKEND_H
#define RT_GFX_BACKEND_H

/* Backend functions and types. */

#include <stddef.h>

#include "gfx.h"

#include "runtime/resources.h"
#include "runtime/rt_math.h"
#include "runtime/runtime.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Handles for backend objects */

#define RT_RENDER_BACKEND_HANDLE_MAX_VERSION 255

#define RT_RENDER_BACKEND_HANDLE(name)                                                             \
    typedef struct {                                                                               \
        uint32_t version : 8;                                                                      \
        uint32_t index : 24;                                                                       \
    } name

RT_RENDER_BACKEND_HANDLE(rt_pipeline_handle);
RT_RENDER_BACKEND_HANDLE(rt_render_target_handle);
RT_RENDER_BACKEND_HANDLE(rt_command_buffer_handle);
RT_RENDER_BACKEND_HANDLE(rt_gpu_semaphore_handle);

#undef RT_RENDER_BACKEND_HANDLE

/* Init data for the renderer */

#ifdef _WIN32
struct HINSTANCE__;
struct HWND__;
#elif defined(RT_USE_XLIB)
struct _XDisplay;
#endif

struct rt_renderer_init_info_s {
#ifdef _WIN32
    struct HINSTANCE__ *hInstance;
    struct HWND__ *hWnd;
#elif defined(RT_USE_XLIB)
    struct _XDisplay *display;
    unsigned long window;
#endif
};

/* Argument types for render commands */

typedef enum {
    RT_GRAPHICS_QUEUE,
    RT_COMPUTE_QUEUE,
    RT_TRANSFER_QUEUE,
} rt_gpu_queue;

typedef struct {
    rt_resource_id vertex_shader;
    rt_resource_id fragment_shader;
    rt_resource_id compute_shader;

    rt_relptr texture_bindings;
    rt_relptr uniform_bindings;
    rt_relptr storage_bindings;

    uint16_t texture_binding_count;
    uint16_t uniform_binding_count;
    uint16_t storage_binding_count;
} rt_pipeline_info;

/* Attributes are used to bind buffers (or textures) to symbolic values.
 * For example, an attribute might be bound to "CELL_GRID", which would be
 * replaced with the (at the time of the invoke) grid buffer of the current
 * world cell.
 */

typedef enum {
    RT_ATTRIBUTE_VALUE_UNDEFINED,

    RT_ATTRIBUTE_VALUE_MATERIAL_ALBEDO,
    RT_ATTRIBUTE_VALUE_MATERIAL_NORMAL,

    RT_ATTRIBUTE_VALUE_count
} rt_attribute_value;

typedef struct {
    uint32_t index;
    rt_attribute_value value;
} rt_attribute_binding;

typedef enum {
    RT_SHADER_TYPE_INVALID,
    RT_SHADER_TYPE_VULKAN,

    RT_SHADER_TYPE_count,
} rt_shader_type;

typedef enum {
    RT_SHADER_STAGE_VERTEX,
    RT_SHADER_STAGE_FRAGMENT,
    RT_SHADER_STAGE_COMPUTE,

    RT_SHADER_STAGE_count,
} rt_shader_stage;

typedef struct {
    rt_shader_type type;
    rt_shader_stage stage;
    rt_relptr bytecode;
    size_t bytecode_length;
} rt_shader_info;

typedef struct {
    rt_gpu_queue target_queue;
} rt_alloc_command_buffer_info;

typedef struct {
    const rt_command_buffer_handle *command_buffers;
    const rt_gpu_semaphore_handle *wait_semaphores;
    const uint64_t *wait_values;
    const rt_gpu_semaphore_handle *signal_semaphores;
    const uint64_t *signal_values;
    uint32_t command_buffer_count;
    uint32_t wait_semaphore_count;
    uint32_t signal_semaphore_count;
} rt_submit_command_buffers_info;

typedef struct {
    /* Optional, for debug purposes */
    const char *name;
    uint64_t initial_value;
} rt_gpu_semaphore_info;

typedef enum {
    RT_PASS_LOAD_MODE_LOAD,
    RT_PASS_LOAD_MODE_CLEAR,
} rt_pass_load_mode;

typedef enum {
    RT_PASS_WRITE_MODE_STORE,
    RT_PASS_WRITE_MODE_DISCARD,
} rt_pass_write_mode;

typedef union {
    rt_color color;
    struct {
        float depth;
        int32_t stencil;
    } depth_stencil;
} rt_pass_clear_value;

typedef struct {
    rt_render_target_handle color_buffers[4];
    rt_pass_load_mode color_buffer_loads[4];
    rt_pass_write_mode color_buffer_writes[4];
    rt_pass_clear_value color_buffer_clear_values[4];
    uint32_t color_buffer_count;

    rt_render_target_handle depth_stencil_buffer;
    rt_pass_load_mode depth_stencil_buffer_load;
    rt_pass_write_mode depth_stencil_buffer_write;
    rt_pass_clear_value depth_stencil_buffer_clear_value;

    rt_rect2i render_area;

    // For debug purposes, can be NULL
    const char *name;
} rt_cmd_begin_pass_info;

typedef enum {
    /* Unusable, must be transitioned to an usable state first. */
    RT_RENDER_TARGET_STATE_INVALID,

    /* Used as a color- or depth-buffer */
    RT_RENDER_TARGET_STATE_ATTACHMENT,

    RT_RENDER_TARGET_STATE_SAMPLED_IMAGE,

    RT_RENDER_TARGET_STATE_STORAGE_IMAGE,
} rt_render_target_state;

/* Renderer API */

typedef void rt_register_renderer_cvars_fn(void);
typedef rt_result rt_init_renderer_fn(const rt_renderer_init_info *info);
typedef void rt_shutdown_renderer_fn(void);
typedef void rt_begin_frame_fn(unsigned int frame_id);
typedef void rt_end_frame_fn(unsigned int frame_id);
typedef rt_pipeline_handle rt_compile_pipeline_fn(const rt_pipeline_info *info);
typedef void rt_destroy_pipeline_fn(rt_pipeline_handle handle);
typedef rt_render_target_handle rt_create_render_target_fn(const rt_render_target_info *info);
typedef rt_render_target_handle rt_get_swapchain_render_target_fn(void);
typedef void rt_destroy_render_target_fn(rt_render_target_handle handle);
typedef rt_result rt_alloc_command_buffers_fn(uint32_t count,
                                              const rt_alloc_command_buffer_info *info,
                                              rt_command_buffer_handle *p_command_buffers);
typedef rt_result rt_submit_command_buffers_fn(rt_gpu_queue queue,
                                               const rt_submit_command_buffers_info *info);
typedef rt_result rt_create_gpu_semaphores_fn(uint32_t count,
                                              const rt_gpu_semaphore_info *info,
                                              rt_gpu_semaphore_handle *p_semaphores);
typedef void rt_destroy_gpu_semaphores_fn(uint32_t count, rt_gpu_semaphore_handle *semaphores);
typedef uint64_t rt_get_gpu_semaphore_value_fn(rt_gpu_semaphore_handle semaphore);
typedef rt_gpu_semaphore_handle rt_get_swapchain_available_semaphore_fn(void);
typedef rt_gpu_semaphore_handle rt_get_render_finished_semaphore_fn(void);

typedef void rt_cmd_begin_pass_fn(rt_command_buffer_handle cmdbuf,
                                  const rt_cmd_begin_pass_info *info);
typedef void rt_cmd_end_pass_fn(rt_command_buffer_handle cmdbuf);
typedef void rt_cmd_transition_render_target_fn(rt_command_buffer_handle cmdbuf,
                                                rt_render_target_handle render_target,
                                                rt_render_target_state new_state);

typedef struct {
    rt_register_renderer_cvars_fn *RegisterCVars;
    rt_init_renderer_fn *Init;
    rt_shutdown_renderer_fn *Shutdown;
    rt_begin_frame_fn *BeginFrame;
    rt_end_frame_fn *EndFrame;
    rt_compile_pipeline_fn *CompilePipeline;
    rt_destroy_pipeline_fn *DestroyPipeline;
    rt_create_render_target_fn *CreateRenderTarget;
    rt_get_swapchain_render_target_fn *GetSwapchainRenderTarget;
    rt_destroy_render_target_fn *DestroyRenderTarget;
    rt_alloc_command_buffers_fn *AllocCommandBuffers;
    rt_submit_command_buffers_fn *SubmitCommandBuffers;
    rt_create_gpu_semaphores_fn *CreateSemaphores;
    rt_destroy_gpu_semaphores_fn *DestroySemaphores;
    rt_get_gpu_semaphore_value_fn *GetSemaphoreValue;
    rt_get_swapchain_available_semaphore_fn *GetSwapchainAvailableSemaphore;
    rt_get_render_finished_semaphore_fn *GetRenderFinishedSemaphore;

    /* Command Buffer Functions */
    rt_cmd_begin_pass_fn *CmdBeginPass;
    rt_cmd_end_pass_fn *CmdEndPass;
    rt_cmd_transition_render_target_fn *CmdTransitionRenderTarget;
} rt_renderer_api;

#define RT_RENDERER_API_FN(name) RT_DLLEXPORT rtRen##name

#ifndef RT_DONT_DEFINE_RENDERER_GLOBAL
extern rt_renderer_api g_renderer;
#endif

#ifdef __cplusplus
}
#endif

#endif
