#include "command_buffers.h"
#include "gpu.h"
#include "render_targets.h"
#include "swapchain.h"

#include "runtime/handles.h"
#include "runtime/mem_arena.h"
#include "gfx/renderer_api.h"

/* Retrieve the VkCommandBuffer as varname, or return */
#define GET_CMDBUF(varname, handle)                                                                \
    VkCommandBuffer varname = rtGetCommandBuffer((handle));                                        \
    if (varname == VK_NULL_HANDLE) {                                                               \
        rtLog("vk", "Failed to retrive VkCommandBuffer for %s", __FUNCTION__);                     \
        return;                                                                                    \
    }

void RT_RENDERER_API_FN(CmdBeginPass)(rt_command_buffer_handle cmdbuf_handle,
                                      const rt_cmd_begin_pass_info *info) {
    GET_CMDBUF(cmdbuf, cmdbuf_handle)

    rt_temp_arena temp = rtGetTemporaryArena(NULL, 0);
    if (!temp.arena) {
        rtReportError("vk", "Failed to acquire a temporary arena for CmdBeginPass");
        return;
    }

#ifdef RT_DEBUG
    VkDebugUtilsLabelEXT debug_label = {
        .sType      = VK_STRUCTURE_TYPE_DEBUG_UTILS_LABEL_EXT,
        .color      = {0.39f, 0.58f, 0.92f, 1.f},
        .pLabelName = (info->name) ? info->name : "Unnamed pass",
    };
    vkCmdBeginDebugUtilsLabelEXT(cmdbuf, &debug_label);
#endif

    /* Acquire the necessary attachments */

    VkRenderingAttachmentInfo *colorbuffers =
        RT_ARENA_PUSH_ARRAY_ZERO(temp.arena, VkRenderingAttachmentInfo, info->color_buffer_count);
    for (uint32_t i = 0; i < info->color_buffer_count; ++i) {
        VkImageView image_view = VK_NULL_HANDLE;
        if (RT_IS_HANDLE_VALID(info->color_buffers[i])) {
            rt_render_target *rt = rtGetRenderTarget(info->color_buffers[i]);
            if (rt)
                image_view = rt->view[g_gpu.current_frame_id % g_gpu.max_frames_in_flight];
        }

        colorbuffers[i].sType       = VK_STRUCTURE_TYPE_RENDERING_ATTACHMENT_INFO;
        colorbuffers[i].pNext       = NULL;
        colorbuffers[i].imageView   = image_view;
        colorbuffers[i].imageLayout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
        switch (info->color_buffer_loads[i]) {
        case RT_PASS_LOAD_MODE_CLEAR:
            colorbuffers[i].loadOp = VK_ATTACHMENT_LOAD_OP_CLEAR;
            break;
        case RT_PASS_LOAD_MODE_LOAD:
            colorbuffers[i].loadOp = VK_ATTACHMENT_LOAD_OP_LOAD;
            break;
        default:
            colorbuffers[i].loadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
            break;
        }
        switch (info->color_buffer_writes[i]) {
        case RT_PASS_WRITE_MODE_STORE:
            colorbuffers[i].storeOp = VK_ATTACHMENT_STORE_OP_STORE;
            break;
        case RT_PASS_WRITE_MODE_DISCARD:
            colorbuffers[i].storeOp = VK_ATTACHMENT_STORE_OP_DONT_CARE;
            break;
        default:
            colorbuffers[i].storeOp = VK_ATTACHMENT_STORE_OP_NONE;
            break;
        }
        memcpy(&colorbuffers[i].clearValue.color.float32,
               info->color_buffer_clear_values[i].color.v,
               sizeof(float) * 4);

        /* TODO: Multisample resolve */
        colorbuffers[i].resolveMode        = VK_RESOLVE_MODE_NONE;
        colorbuffers[i].resolveImageView   = VK_NULL_HANDLE;
        colorbuffers[i].resolveImageLayout = VK_IMAGE_LAYOUT_UNDEFINED;
    }

    /* depth and stencil might be the same */
    VkRenderingAttachmentInfo *depth_stencil_buffer =
        RT_IS_HANDLE_VALID(info->depth_stencil_buffer)
            ? RT_ARENA_PUSH_STRUCT_ZERO(temp.arena, VkRenderingAttachmentInfo)
            : NULL;
    if (depth_stencil_buffer) {
        VkImageView image_view = VK_NULL_HANDLE;
        rt_render_target *rt   = rtGetRenderTarget(info->depth_stencil_buffer);
        if (rt)
            image_view = rt->view[g_gpu.current_frame_id % g_gpu.max_frames_in_flight];

        depth_stencil_buffer->sType       = VK_STRUCTURE_TYPE_RENDERING_ATTACHMENT_INFO;
        depth_stencil_buffer->pNext       = NULL;
        depth_stencil_buffer->imageView   = image_view;
        depth_stencil_buffer->imageLayout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
        switch (info->depth_stencil_buffer_load) {
        case RT_PASS_LOAD_MODE_CLEAR:
            depth_stencil_buffer->loadOp = VK_ATTACHMENT_LOAD_OP_CLEAR;
            break;
        case RT_PASS_LOAD_MODE_LOAD:
            depth_stencil_buffer->loadOp = VK_ATTACHMENT_LOAD_OP_LOAD;
            break;
        default:
            depth_stencil_buffer->loadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
            break;
        }
        switch (info->depth_stencil_buffer_write) {
        case RT_PASS_WRITE_MODE_STORE:
            depth_stencil_buffer->storeOp = VK_ATTACHMENT_STORE_OP_STORE;
            break;
        case RT_PASS_WRITE_MODE_DISCARD:
            depth_stencil_buffer->storeOp = VK_ATTACHMENT_STORE_OP_DONT_CARE;
            break;
        default:
            depth_stencil_buffer->storeOp = VK_ATTACHMENT_STORE_OP_NONE;
            break;
        }

        /* TODO: Multisample resolve */
        depth_stencil_buffer->resolveMode        = VK_RESOLVE_MODE_NONE;
        depth_stencil_buffer->resolveImageView   = VK_NULL_HANDLE;
        depth_stencil_buffer->resolveImageLayout = VK_IMAGE_LAYOUT_UNDEFINED;
    }

    VkRect2D render_area = {
        .offset = {  .x = info->render_area.offset.x,    .y = info->render_area.offset.y},
        .extent = {.width = info->render_area.size.x, .height = info->render_area.size.y}
    };
    if (render_area.extent.width == 0)
        render_area.extent.width = g_swapchain.extent.width;
    if (render_area.extent.height == 0)
        render_area.extent.height = g_swapchain.extent.height;

    VkRenderingInfo rendering_info = {
        .sType                = VK_STRUCTURE_TYPE_RENDERING_INFO,
        .pColorAttachments    = colorbuffers,
        .colorAttachmentCount = info->color_buffer_count,
        .pDepthAttachment     = depth_stencil_buffer,
        .pStencilAttachment   = depth_stencil_buffer,
        .layerCount           = 1,
        .renderArea           = render_area,
    };

    vkCmdBeginRendering(cmdbuf, &rendering_info);

    rtReturnTemporaryArena(temp);
}

void RT_RENDERER_API_FN(CmdEndPass)(rt_command_buffer_handle cmdbuf_handle) {
    GET_CMDBUF(cmdbuf, cmdbuf_handle)
    vkCmdEndRendering(cmdbuf);

#ifdef RT_DEBUG
    vkCmdEndDebugUtilsLabelEXT(cmdbuf);
#endif
}

void RT_RENDERER_API_FN(CmdTransitionRenderTarget)(rt_command_buffer_handle cmdbuf_handle,
                                                   rt_render_target_handle render_target,
                                                   rt_render_target_state new_state) {
    GET_CMDBUF(cmdbuf, cmdbuf_handle)
    uint32_t image_index = g_gpu.current_frame_id % g_gpu.max_frames_in_flight;
    if (render_target.index == g_renderer.GetSwapchainRenderTarget().index) {
        image_index = rtGetFrameData(g_gpu.current_frame_id)->swapchain_image_index;
    }

    rt_render_target *rt = rtGetRenderTarget(render_target);
    if (!rt) {
        rtLog("vk", "Tried to transition invalid render target");
        return;
    }

    if (rt->states[image_index] == new_state)
        return;

    /* Determine old and new layout */
    VkImageLayout old_layout;
    switch (rt->states[image_index]) {
    case RT_RENDER_TARGET_STATE_ATTACHMENT:
        if (rt->format == VK_FORMAT_D24_UNORM_S8_UINT || rt->format == VK_FORMAT_D32_SFLOAT) {
            old_layout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
        } else {
            old_layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
        }
        break;
    case RT_RENDER_TARGET_STATE_STORAGE_IMAGE:
    case RT_RENDER_TARGET_STATE_SAMPLED_IMAGE:
        old_layout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        break;
    default:
        old_layout = VK_IMAGE_LAYOUT_UNDEFINED;
        break;
    }
    VkImageLayout new_layout;
    switch (new_state) {
    case RT_RENDER_TARGET_STATE_ATTACHMENT:
        if (rt->format == VK_FORMAT_D24_UNORM_S8_UINT || rt->format == VK_FORMAT_D32_SFLOAT) {
            new_layout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
        } else {
            new_layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
        }
        break;
    case RT_RENDER_TARGET_STATE_STORAGE_IMAGE:
    case RT_RENDER_TARGET_STATE_SAMPLED_IMAGE:
        new_layout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        break;
    default:
        new_layout = VK_IMAGE_LAYOUT_UNDEFINED;
    }

#ifdef RT_DEBUG
    VkDebugUtilsLabelEXT debug_label = {
        .sType      = VK_STRUCTURE_TYPE_DEBUG_UTILS_LABEL_EXT,
        .pLabelName = "Transition Render Target",
        .color      = {.13f, .54f, .13f, .75f},
    };
    vkCmdBeginDebugUtilsLabelEXT(cmdbuf, &debug_label);
#endif

    VkImageAspectFlags aspect_mask =
        (rt->format == VK_FORMAT_D24_UNORM_S8_UINT || rt->format == VK_FORMAT_D32_SFLOAT)
            ? VK_IMAGE_ASPECT_DEPTH_BIT | VK_IMAGE_ASPECT_STENCIL_BIT
            : VK_IMAGE_ASPECT_COLOR_BIT;
    VkImageMemoryBarrier2 image_barrier = {
        .sType         = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER_2,
        .srcStageMask  = VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT,
        .srcAccessMask = VK_ACCESS_2_MEMORY_WRITE_BIT,
        .dstStageMask  = VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT,
        .dstAccessMask = VK_ACCESS_2_MEMORY_WRITE_BIT | VK_ACCESS_2_MEMORY_READ_BIT,
        .oldLayout     = old_layout,
        .newLayout     = new_layout,
        .image         = rt->image[image_index],
 /* clang-format off */
        .subresourceRange = {
            .aspectMask = aspect_mask,
            .baseArrayLayer = 0,
            .baseMipLevel = 0,
            .layerCount = 1,
            .levelCount = 1,
        },
  /* clang-format on */
    };

    VkDependencyInfo dep_info = {
        .sType                   = VK_STRUCTURE_TYPE_DEPENDENCY_INFO,
        .pImageMemoryBarriers    = &image_barrier,
        .imageMemoryBarrierCount = 1,
    };
    vkCmdPipelineBarrier2(cmdbuf, &dep_info);

#ifdef RT_DEBUG
    vkCmdEndDebugUtilsLabelEXT(cmdbuf);
#endif

    rt->states[image_index] = new_state;
}
