#include "command_buffers.h"
#include "gpu.h"
#include "render_targets.h"
#include "swapchain.h"

#include "gfx/renderer_api.h"

#define ONE_SECOND_NS 1000000000u

void RT_RENDERER_API_FN(BeginFrame)(unsigned int frame_id) {
    g_gpu.current_frame_id = frame_id;

    rt_frame_data *frame = rtGetFrameData(frame_id);

    /* Wait until the previous frame is done */
    VkFence fence = g_swapchain.image_fences[frame_id % g_swapchain.image_count];
    RT_VK_CHECK(vkWaitForFences(g_gpu.device, 1, &fence, VK_TRUE, ONE_SECOND_NS));
    RT_VK_CHECK(vkResetFences(g_gpu.device, 1, &fence));

    rtResetCommandPools(frame_id);

    VkResult acquire_res = vkAcquireNextImageKHR(g_gpu.device,
                                                 g_swapchain.swapchain,
                                                 ONE_SECOND_NS,
                                                 frame->image_available,
                                                 fence,
                                                 &frame->swapchain_image_index);
    if (acquire_res == VK_SUBOPTIMAL_KHR || acquire_res == VK_ERROR_OUT_OF_DATE_KHR) {
        /* We need to recreate the swapchain and try again */
        rtLog("vk", "Swapchain has become suboptimal and needs to be re-created.");
        vkDeviceWaitIdle(g_gpu.device);
        if (rtRecreateSwapchain() != RT_SUCCESS) {
            rtReportError("vk", "Failed to recreate the swapchain.");
            return;
        }
        rtUpdateSwapchainRenderTarget();
        rtUpdateRenderTargetsFromSwapchain(g_swapchain.image_count,
                                           g_swapchain.format,
                                           g_swapchain.extent);
        rtRenBeginFrame(frame_id);
    } else if (acquire_res != VK_SUCCESS) {
        rtReportError("vk", "vkAcquireNextImageKHR failed: %u", acquire_res);
    }

    /* Update the swapchain render target */
    rt_render_target_handle swap_rt_handle        = g_renderer.GetSwapchainRenderTarget();
    rt_render_target *swap_rt                     = rtGetRenderTarget(swap_rt_handle);
    swap_rt->states[frame->swapchain_image_index] = RT_RENDER_TARGET_STATE_INVALID;
}

void RT_RENDERER_API_FN(EndFrame)(unsigned int frame_id) {
    rt_frame_data *frame = rtGetFrameData(frame_id);

    uint32_t image_index = frame->swapchain_image_index;

    /* Transition the swap chain image to the correct layout */
    VkCommandBuffer cmd = rtAllocSingleCommandBuffer(RT_GRAPHICS_QUEUE);
    if (cmd == VK_NULL_HANDLE) {
        rtReportError("vk",
                      "Failed to allocate a command buffer for transitioning the swapchain image "
                      "to PRESENT_SRC layout.");
        return;
    }

    VkCommandBufferBeginInfo begin_info = {.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
                                           .flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT};
    vkBeginCommandBuffer(cmd, &begin_info);

#ifdef RT_DEBUG
    VkDebugUtilsLabelEXT debug_label = {
        .sType      = VK_STRUCTURE_TYPE_DEBUG_UTILS_LABEL_EXT,
        .color      = {.13f, .54f, .13f, 1.f},
        .pLabelName = "Transition Swapchain"
    };
    vkCmdBeginDebugUtilsLabelEXT(cmd, &debug_label);
#endif

    VkImageMemoryBarrier2 image_barrier = {
        .sType         = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER_2,
        .srcStageMask  = VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT,
        .srcAccessMask = VK_ACCESS_2_MEMORY_WRITE_BIT,
        .dstStageMask  = VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT,
        .dstAccessMask = VK_ACCESS_2_MEMORY_WRITE_BIT | VK_ACCESS_2_MEMORY_READ_BIT,
        .oldLayout     = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL,
        .newLayout     = VK_IMAGE_LAYOUT_PRESENT_SRC_KHR,
        .image         = g_swapchain.images[image_index],
 /* clang-format off */
        .subresourceRange = {
            .aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
            .baseArrayLayer = 0,
            .baseMipLevel = 0,
            .layerCount = 1,
            .levelCount = 1,
        },
  /* clang-format on */
    };

    VkDependencyInfo dep_info = {
        .sType                   = VK_STRUCTURE_TYPE_DEPENDENCY_INFO,
        .pImageMemoryBarriers    = &image_barrier,
        .imageMemoryBarrierCount = 1,
    };
    vkCmdPipelineBarrier2(cmd, &dep_info);

#ifdef RT_DEBUG
    vkCmdEndDebugUtilsLabelEXT(cmd);
#endif

    vkEndCommandBuffer(cmd);
    if (rtSubmitSingleCommandBuffer(cmd,
                                    &frame->render_finished,
                                    NULL,
                                    1,
                                    &frame->swapchain_transitioned,
                                    NULL,
                                    1,
                                    RT_GRAPHICS_QUEUE) != RT_SUCCESS) {
        rtReportError("vk", "Failed to submit the layout transition for the swapchain image.");
        return;
    }

    VkPresentInfoKHR present_info = {
        .sType              = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,
        .pImageIndices      = &image_index,
        .pSwapchains        = &g_swapchain.swapchain,
        .swapchainCount     = 1,
        .pWaitSemaphores    = &frame->swapchain_transitioned,
        .waitSemaphoreCount = 1,
    };

    VkResult res = vkQueuePresentKHR(g_gpu.present_queue, &present_info);
    if (res != VK_SUCCESS) {
        rtReportError("vk", "vkQueuePresentKHR failed: %u", res);
    }
}