#include "runtime/config.h"
#include "runtime/threading.h"

#include "gfx/renderer_api.h"

#include "gpu.h"
#include "render_targets.h"
#include "swapchain.h"

#include <stdlib.h>
#include <volk/volk.h>

RT_CVAR_I(r_VkMaxRenderTargetCount, "Maximum number of render target objects. Default: 1024", 1024);

typedef struct rt_render_target_slot_s {
    uint32_t version;
    rt_render_target render_target;
    struct rt_render_target_slot_s *next_free;
} rt_render_target_slot;

static rt_render_target_slot *_render_targets;
static rt_render_target_slot *_first_free;
static rt_rwlock _lock;

static rt_render_target_handle _swapchain_handle;

static void DestroyRenderTarget(rt_render_target_slot *slot) {
    for (unsigned int i = 0; i < slot->render_target.image_count; ++i) {
        vkDestroyImageView(g_gpu.device, slot->render_target.view[i], g_gpu.alloc_cb);
        vmaDestroyImage(g_gpu.allocator,
                        slot->render_target.image[i],
                        slot->render_target.allocation[i]);
    }
    slot->next_free = _first_free;
    _first_free     = slot;
}

static bool CreateImageAndView(VkExtent2D extent,
                               VkFormat format,
                               VkSampleCountFlagBits sample_count,
                               VkImageUsageFlagBits usage,
                               VkImageAspectFlagBits aspect,
                               VkImage *p_image,
                               VmaAllocation *p_allocation,
                               VkImageView *p_view) {

    uint32_t queue_families[3];
    uint32_t distinct_queue_families = 1;
    queue_families[0]                = g_gpu.graphics_family;
    if (g_gpu.compute_family != g_gpu.graphics_family)
        queue_families[distinct_queue_families++] = g_gpu.compute_family;
    if (g_gpu.present_family != g_gpu.graphics_family &&
        g_gpu.present_family != g_gpu.compute_family)
        queue_families[distinct_queue_families++] = g_gpu.present_family;

    VkFormatProperties2 props = {
        .sType = VK_STRUCTURE_TYPE_FORMAT_PROPERTIES_2,
    };
    vkGetPhysicalDeviceFormatProperties2(g_gpu.phys_device, format, &props);
    if ((props.formatProperties.optimalTilingFeatures & VK_FORMAT_FEATURE_SAMPLED_IMAGE_BIT) == 0) {
        rtLog("vk",
              "Requested render target format %s can not be sampled.",
              rtVkFormatToString(format));
        usage &= ~VK_IMAGE_USAGE_SAMPLED_BIT;
    }
    if ((props.formatProperties.optimalTilingFeatures & VK_FORMAT_FEATURE_STORAGE_IMAGE_BIT) == 0) {
        rtLog("vk",
              "Requested render target format %s can not be used for storage.",
              rtVkFormatToString(format));
        usage &= ~VK_IMAGE_USAGE_STORAGE_BIT;
    }
    if (((usage & VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT) != 0) &&
        ((props.formatProperties.optimalTilingFeatures & VK_FORMAT_FEATURE_COLOR_ATTACHMENT_BIT) ==
         0)) {
        rtReportError(
            "vk",
            "Tried to create a render target color attachment, but the format %s does not "
            "support the color attachment usage.",
            rtVkFormatToString(format));
        return false;
    } else if (((usage & VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT) != 0) &&
               ((props.formatProperties.optimalTilingFeatures &
                 VK_FORMAT_FEATURE_DEPTH_STENCIL_ATTACHMENT_BIT) == 0)) {
        rtReportError("vk",
                      "Tried to create a render target depth/stencil attachment, but the format %s"
                      "does not support the depth/stencil attachment usage.",
                      rtVkFormatToString(format));
        return false;
    }

    VkImageCreateInfo image_info = {
        .sType       = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO,
        .imageType   = VK_IMAGE_TYPE_2D,
        .format      = format,
        .extent      = {.width = extent.width, .height = extent.height, .depth = 1},
        .mipLevels   = 1,
        .arrayLayers = 1,
        .samples     = sample_count,
        .tiling      = VK_IMAGE_TILING_OPTIMAL,
        .usage       = usage,
        .sharingMode =
            (distinct_queue_families > 1) ? VK_SHARING_MODE_CONCURRENT : VK_SHARING_MODE_EXCLUSIVE,
        .pQueueFamilyIndices   = (distinct_queue_families > 1) ? queue_families : NULL,
        .queueFamilyIndexCount = distinct_queue_families,
    };
    VmaAllocationCreateInfo alloc_info = {
        .usage = VMA_MEMORY_USAGE_GPU_ONLY,
    };

    VkImage image;
    VmaAllocation allocation;
    if (vmaCreateImage(g_gpu.allocator, &image_info, &alloc_info, &image, &allocation, NULL) !=
        VK_SUCCESS) {
        return false;
    }

    VkImageViewCreateInfo view_info = {
        .sType      = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO,
        .image      = image,
        .viewType   = VK_IMAGE_VIEW_TYPE_2D,
        .format     = format,
        .components = {.r = VK_COMPONENT_SWIZZLE_IDENTITY,
                       .g = VK_COMPONENT_SWIZZLE_IDENTITY,
                       .b = VK_COMPONENT_SWIZZLE_IDENTITY,
                       .a = VK_COMPONENT_SWIZZLE_IDENTITY},
 /* clang-format off */
            .subresourceRange = {
                           .aspectMask     = aspect,
                           .baseArrayLayer = 0,
                           .baseMipLevel   = 0,
                           .layerCount     = 1,               
                           .levelCount     = 1,
            },
  /* clang-format on */
    };
    VkImageView view;
    if (vkCreateImageView(g_gpu.device, &view_info, g_gpu.alloc_cb, &view) != VK_SUCCESS) {
        rtLog("VK", "Failed to create render target image view");
        vmaDestroyImage(g_gpu.allocator, image, allocation);
        return false;
    }

    *p_image      = image;
    *p_allocation = allocation;
    *p_view       = view;
    return true;
}

rt_result InitRenderTargetManagement(void) {
    rt_create_rwlock_result lock_res = rtCreateRWLock();
    if (!lock_res.ok)
        return RT_UNKNOWN_ERROR;
    _lock = lock_res.lock;

    _render_targets = calloc(r_VkMaxRenderTargetCount.i, sizeof(rt_render_target_slot));
    if (!_render_targets) {
        rtDestroyRWLock(&_lock);
        return RT_OUT_OF_MEMORY;
    }
    /* Keep [0] unused to preserve 0 as the invalid handle */
    _first_free = &_render_targets[1];
    for (int i = 1; i < r_VkMaxRenderTargetCount.i - 1; ++i) {
        _render_targets[i].next_free = &_render_targets[i + 1];
    }

    /* Reserve the slot for the swap chain rt */
    rt_render_target_slot *slot = _first_free;
    _first_free                 = slot->next_free;
    slot->version               = (slot->version + 1) & RT_RENDER_BACKEND_HANDLE_MAX_VERSION;
    _swapchain_handle           = (rt_render_target_handle){.version = slot->version,
                                                            .index   = (uint32_t)(slot - _render_targets)};

    return RT_SUCCESS;
}

void ShutdownRenderTargetManagement(void) {
    for (int i = 1; i < r_VkMaxRenderTargetCount.i; ++i) {
        DestroyRenderTarget(&_render_targets[i]);
    }
    free(_render_targets);
    rtDestroyRWLock(&_lock);
    _first_free = NULL;
}

rt_render_target_handle RT_RENDERER_API_FN(CreateRenderTarget)(const rt_render_target_info *info) {
    rt_render_target_handle handle = {0};

    rtLockWrite(&_lock);
    if (!_first_free) {
        rtLog("VK", "No free render target slots!");
        rtUnlockWrite(&_lock);
        return handle;
    }
    rt_render_target_slot *slot = _first_free;
    _first_free                 = slot->next_free;
    slot->version               = (slot->version + 1) & RT_RENDER_BACKEND_HANDLE_MAX_VERSION;

    /* No other thread that calls compile gets the same slot.
     * Another thread accessing the slot via GetPipeline would get a version mismatch.
     * The same holds for DestroyPipeline
     */
    rtUnlockWrite(&_lock);

    slot->render_target.match_swapchain = 0;
    slot->render_target.image_count     = g_swapchain.image_count;
    for (unsigned int i = 0; i < g_swapchain.image_count; ++i) {
        uint32_t width = info->width, height = info->height;
        if (width == RT_RENDER_TARGET_SIZE_SWAPCHAIN) {
            width = g_swapchain.extent.width;
            slot->render_target.match_swapchain |= RT_RENDER_TARGET_MATCH_SWAPCHAIN_SIZE;
        }
        if (height == RT_RENDER_TARGET_SIZE_SWAPCHAIN) {
            height = g_swapchain.extent.height;
            slot->render_target.match_swapchain |= RT_RENDER_TARGET_MATCH_SWAPCHAIN_SIZE;
        }
        slot->render_target.extent = (VkExtent2D){.width = width, .height = height};

        if (info->format != RT_PIXEL_FORMAT_SWAPCHAIN)
            slot->render_target.format = rtPixelFormatToVkFormat(info->format);
        else {
            slot->render_target.format = g_swapchain.format;
            slot->render_target.match_swapchain |= RT_RENDER_TARGET_MATCH_SWAPCHAIN_FORMAT;
        }
        if (info->format == RT_PIXEL_FORMAT_DEPTH24_STENCIL8 ||
            info->format == RT_PIXEL_FORMAT_DEPTH32) {
            slot->render_target.usage = VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT |
                                        VK_IMAGE_USAGE_SAMPLED_BIT | VK_IMAGE_USAGE_STORAGE_BIT;
            if (info->format == RT_PIXEL_FORMAT_DEPTH32)
                slot->render_target.aspect = VK_IMAGE_ASPECT_DEPTH_BIT;
            else
                slot->render_target.aspect =
                    VK_IMAGE_ASPECT_DEPTH_BIT | VK_IMAGE_ASPECT_STENCIL_BIT;
        } else {
            slot->render_target.usage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT |
                                        VK_IMAGE_USAGE_SAMPLED_BIT | VK_IMAGE_USAGE_STORAGE_BIT;
            slot->render_target.aspect = VK_IMAGE_ASPECT_COLOR_BIT;
        }
        slot->render_target.sample_count = rtSampleCountToFlags(info->sample_count);
        if (!CreateImageAndView(slot->render_target.extent,
                                slot->render_target.format,
                                slot->render_target.sample_count,
                                slot->render_target.usage,
                                slot->render_target.aspect,
                                &slot->render_target.image[i],
                                &slot->render_target.allocation[i],
                                &slot->render_target.view[i])) {
            slot->render_target.image_count = i;
            DestroyRenderTarget(slot);
            goto out;
        }
        slot->render_target.states[i] = RT_RENDER_TARGET_STATE_INVALID;
    }

    handle.version = slot->version;
    handle.index   = (uint32_t)(slot - _render_targets);
out:
    return handle;
}

rt_render_target_handle RT_RENDERER_API_FN(GetSwapchainRenderTarget)(void) {
    return _swapchain_handle;
}

void RT_RENDERER_API_FN(DestroyRenderTarget)(rt_render_target_handle handle) {
    if (handle.index >= (uint32_t)r_VkMaxRenderTargetCount.i)
        return;
    rtLockWrite(&_lock);
    if (_render_targets[handle.index].version == handle.version)
        DestroyRenderTarget(&_render_targets[handle.index]);
    else
        rtLog("VK", "Tried to destroy a render target using an outdated handle.");
    rtUnlockWrite(&_lock);
}

rt_render_target *rtGetRenderTarget(rt_render_target_handle handle) {
    if (handle.index >= (uint32_t)r_VkMaxRenderTargetCount.i)
        return NULL;
    rtLockRead(&_lock);
    rt_render_target *res = NULL;
    if (_render_targets[handle.index].version == handle.version)
        res = &_render_targets[handle.index].render_target;
    else
        rtLog("VK", "Tried to access a render target using an outdated handle.");
    rtUnlockRead(&_lock);
    return res;
}

void rtUpdateSwapchainRenderTarget(void) {
    RT_ASSERT(_swapchain_handle.index != 0, "Invalid swap chain render target!");
    rt_render_target_slot *slot = &_render_targets[_swapchain_handle.index];
    rt_render_target *rt        = &slot->render_target;

    rt->match_swapchain = 0;
    rt->format          = g_swapchain.format;
    rt->extent          = g_swapchain.extent;
    rt->sample_count    = 1;
    rt->usage           = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;
    rt->aspect          = VK_IMAGE_ASPECT_COLOR_BIT;
    for (uint32_t i = 0; i < g_swapchain.image_count; ++i) {
        rt->allocation[i] = NULL;
        rt->image[i]      = g_swapchain.images[i];
        rt->view[i]       = g_swapchain.image_views[i];
        rt->states[i]     = RT_RENDER_TARGET_STATE_INVALID;
    }
}

void rtUpdateRenderTargetsFromSwapchain(uint32_t image_count, VkFormat format, VkExtent2D extent) {
    rtLockWrite(&_lock);
    for (uint32_t i = 1; i < (uint32_t)r_VkMaxRenderTargetCount.i; ++i) {
        if (_render_targets[i].render_target.image_count == 0)
            continue;

        rt_render_target *render_target = &_render_targets[i].render_target;
        if (render_target->match_swapchain != 0) {
            for (uint32_t j = 0; j < render_target->image_count; ++j) {
                vkDestroyImageView(g_gpu.device, render_target->view[j], g_gpu.alloc_cb);
                vmaDestroyImage(g_gpu.allocator,
                                render_target->image[j],
                                render_target->allocation[j]);
            }
            if ((render_target->match_swapchain & RT_RENDER_TARGET_MATCH_SWAPCHAIN_FORMAT) != 0) {
                render_target->format = format;
            } else if ((render_target->match_swapchain & RT_RENDER_TARGET_MATCH_SWAPCHAIN_SIZE) !=
                       0) {
                render_target->extent = extent;
            }
            for (uint32_t j = 0; j < image_count; ++j) {
                if (!CreateImageAndView(render_target->extent,
                                        render_target->format,
                                        render_target->sample_count,
                                        render_target->usage,
                                        render_target->aspect,
                                        &render_target->image[j],
                                        &render_target->allocation[j],
                                        &render_target->view[j])) {
                    render_target->image_count = j;
                    DestroyRenderTarget(&_render_targets[i]);
                    rtReportError("VK", "Failed to recreate swapchain-matching render target");
                    break;
                }
            }
        } else if (render_target->image_count < image_count) {
            /* Create additional images */
            for (uint32_t j = render_target->image_count; j < image_count; ++j) {
                if (!CreateImageAndView(render_target->extent,
                                        render_target->format,
                                        render_target->sample_count,
                                        render_target->usage,
                                        render_target->aspect,
                                        &render_target->image[j],
                                        &render_target->allocation[j],
                                        &render_target->view[j])) {
                    render_target->image_count = j;
                    DestroyRenderTarget(&_render_targets[i]);
                    rtReportError("VK", "Failed to create additional render target images");
                    break;
                }
            }
        } else if (render_target->image_count > image_count) {
            /* Delete unnecessary images */
            for (uint32_t j = image_count; j < render_target->image_count; ++j) {
                vkDestroyImageView(g_gpu.device, render_target->view[j], g_gpu.alloc_cb);
                vmaDestroyImage(g_gpu.allocator,
                                render_target->image[j],
                                render_target->allocation[j]);
            }
        }
        render_target->image_count = image_count;
    }
    rtUnlockWrite(&_lock);
}